@@ -187,6 +187,15 @@ struct respip_addr_info;
 /** Maximum number of known edns options */
 #define MAX_KNOWN_EDNS_OPTS 256
 
+struct errinf_strlist {
+    /** next item in list */
+    struct errinf_strlist* next;
+    /** config option string */
+    char* str;
+    /** EDE code companion to the error str */
+    int reason_bogus;
+};
+
 enum inplace_cb_list_type {
 	/* Inplace callbacks for when a resolved reply is ready to be sent to the
 	 * front.*/
@@ -624,8 +633,7 @@ struct module_qstate {
 	/** region for this query. Cleared when query process finishes. */
 	struct regional* region;
 	/** failure reason information if val-log-level is high */
-	struct config_strlist* errinf;
-
+	struct errinf_strlist* errinf;
 	/** which module is executing */
 	int curmod;
 	/** module states */
@@ -667,6 +675,8 @@ struct module_qstate {
 	/** Extended result of response-ip action processing, mainly
 	 *  for logging purposes. */
 	struct respip_action_info* respip_action_info;
+	/** if the query is rpz passthru, no further rpz processing for it */
+	int rpz_passthru;
 
 	/** whether the reply should be dropped */
 	int is_drop;
@@ -759,6 +769,65 @@ const char* strextstate(enum module_ext_state s);
  */
 const char* strmodulevent(enum module_ev e);
 
+/**
+ * Append text to the error info for validation.
+ * @param qstate: query state.
+ * @param str: copied into query region and appended.
+ * Failures to allocate are logged.
+ */
+void errinf(struct module_qstate* qstate, const char* str);
+void errinf_ede(struct module_qstate* qstate, const char* str,
+                sldns_ede_code reason_bogus);
+
+/**
+ * Append text to error info:  from 1.2.3.4
+ * @param qstate: query state.
+ * @param origin: sock list with origin of trouble. 
+ *  Every element added.
+ *  If NULL: nothing is added.
+ *  if 0len element: 'from cache' is added.
+ */
+void errinf_origin(struct module_qstate* qstate, struct sock_list *origin);
+
+/**
+ * Append text to error info:  for RRset name type class
+ * @param qstate: query state.
+ * @param rr: rrset_key.
+ */
+void errinf_rrset(struct module_qstate* qstate, struct ub_packed_rrset_key *rr);
+
+/**
+ * Append text to error info:  str dname
+ * @param qstate: query state.
+ * @param str: explanation string
+ * @param dname: the dname.
+ */
+void errinf_dname(struct module_qstate* qstate, const char* str, 
+    uint8_t* dname);
+
+/**
+ * Create error info in string.  For validation failures.
+ * @param qstate: query state.
+ * @return string or NULL on malloc failure (already logged).
+ *    This string is malloced and has to be freed by caller.
+ */
+char* errinf_to_str_bogus(struct module_qstate* qstate);
+/**
+ * Check the sldns_ede_code of the qstate.
+ * @param qstate: query state.
+ * @return LDNS_EDE_DNSSEC_BOGUS by default, or the first explicitly set
+ *   sldns_ede_code.
+ */
+sldns_ede_code errinf_to_reason_bogus(struct module_qstate* qstate);
+
+/**
+ * Create error info in string.  For other servfails.
+ * @param qstate: query state.
+ * @return string or NULL on malloc failure (already logged).
+ *    This string is malloced and has to be freed by caller.
+ */
+char* errinf_to_str_servfail(struct module_qstate* qstate);
+
 /**
  * Initialize the edns known options by allocating the required space.
  * @param env: the module environment.
